"""
An extractor able to resolve single-row lookups from Cassandra tables in
a keyspace, with a fair amount of metadata inspection.
"""

from functools import reduce

from cassandra.query import SimpleStatement

from cassio.inspection import (
    _table_primary_key_columns,
)


class CassandraExtractor():

    def __init__(self, session, keyspace, field_mapper, literal_nones):
        self.session = session
        self.keyspace = keyspace
        self.field_mapper = field_mapper
        self.literal_nones = literal_nones # TODO: handle much better
        # derived fields
        self.tablesNeeded = {fmv[0] for fmv in field_mapper.values()}
        self.primaryKeyMap = {
            tableName: _table_primary_key_columns(self.session, self.keyspace, tableName)
            for tableName in self.tablesNeeded
        }
        # all primary-key values needed across tables
        self.requiredParameters = list(reduce(lambda accum, nw: accum | set(nw), self.primaryKeyMap.values(), set()))
        # TODOs:
        #   move this getter creation someplace else
        #   query a table only once (grouping required variables by source table, selecting only those unless function passed)
        def _getter(**kwargs):
            def _retrieve_field(_tableName2, _keyColumns, _columnOrExtractor, _keyValueMap):
                selector = SimpleStatement('SELECT * FROM {keyspace}.{tableName} WHERE {whereClause} LIMIT 1;'.format(
                    keyspace=keyspace,
                    tableName=_tableName2,
                    whereClause=' AND '.join(
                        f'{kc} = %s'
                        for kc in _keyColumns
                    ),
                ))
                values = tuple([
                    _keyValueMap[kc]
                    for kc in _keyColumns
                ])
                row = session.execute(selector, values).one()
                if row:
                    if callable(_columnOrExtractor):
                        return _columnOrExtractor(row)
                    else:
                        return getattr(row, _columnOrExtractor)
                else:
                    if literal_nones:
                        return None
                    else:
                        raise ValueError('No data found for %s from %s.%s' % (
                            str(_columnOrExtractor),
                            keyspace,
                            _tableName2,
                        ))
            
            return {
                field: _retrieve_field(tableName, self.primaryKeyMap[tableName], columnOrExtractor, kwargs)
                for field, (tableName, columnOrExtractor) in field_mapper.items()
            }
        self.getter = _getter

    def __call__(self, *pargs, **kwargs):
        return self.getter(*pargs, **kwargs)



"""
A common driver to operate on tables with vector-similarity-search indices.
"""

from operator import itemgetter
import json

from cassandra.cluster import Session
from cassandra.query import SimpleStatement
from cassandra import ReadFailure

from cassio.globals.globals import globals
from cassio.utils.vector.distance_metrics import distanceMetricsMap

EXPERIMENTAL_VECTOR_SEARCH_ERROR = (
    "Vector search is an experimental feature and should "
    "be first enabled with 'cassio.globals.enableExperimentalVectorSearch()`"
)

_createVectorDBTableCQLTemplate = """
CREATE TABLE IF NOT EXISTS {keyspace}.{tableName} (
    document_id {idType} PRIMARY KEY,
    embedding_vector FLOAT VECTOR[{embeddingDimension}],
    document TEXT,
    metadata_blob TEXT
);
"""
_createVectorDBTableIndexCQLTemplate = """
CREATE CUSTOM INDEX IF NOT EXISTS {indexName} ON {keyspace}.{tableName} (embedding_vector)
USING 'org.apache.cassandra.index.sai.StorageAttachedIndex' ;
"""
_storeCachedVSSItemCQLTemplate = """
INSERT INTO {keyspace}.{tableName} (
    document_id,
    embedding_vector,
    document,
    metadata_blob
) VALUES (
    {documentIdPlaceholder},
    %s,
    %s,
    %s
){ttlSpec};
"""
_getVectorDBTableItemCQLTemplate = """
SELECT
    document_id, embedding_vector, document, metadata_blob
FROM {keyspace}.{tableName}
    WHERE document_id=%s;
"""
_searchVectorDBTableItemCQLTemplate = """
SELECT
    document_id, embedding_vector, document, metadata_blob
FROM {keyspace}.{tableName}
    WHERE embedding_vector ANN OF %s
    LIMIT %s
"""
_truncateVectorDBTableCQLTemplate = """
TRUNCATE TABLE {keyspace}.{tableName};
"""
_deleteVectorDBTableItemCQLTemplate = """
DELETE FROM {keyspace}.{tableName}
WHERE document_id = %s;
"""
_countRowsCQLTemplate = """
    SELECT COUNT(*) FROM {keyspace}.{tableName};
"""


class VectorDBMixin():

    def _createIndex(self):
        indexName = f'{self.tableName}_embedding_idx'
        createVectorDBTableIndexCQL = SimpleStatement(_createVectorDBTableIndexCQLTemplate.format(
            indexName=indexName,
            keyspace=self.keyspace,
            tableName=self.tableName
        ))
        self._executeCQL(createVectorDBTableIndexCQL, tuple())

    def ANNSearch(self, embedding_vector, numRows):
        """
        Current versions of vector-search Cassandra fail when more rows than
        present in the table are asked in the LIMIT clause.
        The solution below tries to fix that.
        """
        try:
            searchVectorDBTableItemCQL = SimpleStatement(_searchVectorDBTableItemCQLTemplate.format(
                keyspace=self.keyspace,
                tableName=self.tableName
            ))
            return self._executeCQL(searchVectorDBTableItemCQL, (embedding_vector, numRows))
        except ReadFailure:
            # likely a count issue. Let's count the rows (it's a small number)
            rowCount = self._countRows()
            if rowCount < numRows:
                if rowCount:
                    return self.ANNSearch(embedding_vector, min(numRows, rowCount))
                else:
                    return []
            else:
                # a repeated error, likely another reason. Reraise
                raise

    def _countRows(self):
        countRowsCQL = SimpleStatement(_countRowsCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName
        ))
        return self._executeCQL(countRowsCQL, tuple()).one().count


class VectorDBTable(VectorDBMixin):

    def __init__(self, session: Session, keyspace: str, tableName: str, embeddingDimension: int, autoID: bool):
        if not globals.experimentalVectorSearch:
            raise RuntimeError(EXPERIMENTAL_VECTOR_SEARCH_ERROR)
        #
        self.session = session
        self.keyspace = keyspace
        self.tableName = tableName
        self.embeddingDimension = embeddingDimension
        #
        self.autoID = autoID
        #
        self._createTable()
        self._createIndex()

    def put(self, document, embedding_vector, document_id, metadata, ttlSeconds):
        # document_id, if not autoID, must be str
        if not self.autoID and document_id is None:
            raise ValueError('\'document_id\' must be specified unless autoID')
        if self.autoID and document_id is not None:
            raise ValueError('\'document_id\' cannot be passes if autoID')
        if ttlSeconds:
            ttlSpec = f' USING TTL {ttlSeconds}'
        else:
            ttlSpec = ''
        storeCachedVSSItemCQL = SimpleStatement(_storeCachedVSSItemCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
            documentIdPlaceholder='now()' if self.autoID else '%s',
            ttlSpec=ttlSpec,
        ))
        metadataBlob = json.dumps(metadata)
        # depending on autoID, the size of the values tuple changes:
        values0 = (embedding_vector, document, metadataBlob)
        values = values0 if self.autoID else tuple([document_id] + list(values0))
        self._executeCQL(storeCachedVSSItemCQL, values)

    def get(self, document_id):
        if self.autoID:
            raise ValueError('\'get\' not supported if autoID')
        else:
            getVectorDBTableItemCQL = SimpleStatement(_getVectorDBTableItemCQLTemplate.format(
                keyspace=self.keyspace,
                tableName=self.tableName,
            ))
            hits = self._executeCQL(getVectorDBTableItemCQL, (document_id, ))
            hit = hits.one()
            if hit:
                return VectorDBTable._jsonifyHit(hit, distance=None)
            else:
                return None

    def delete(self, document_id) -> None:
        """This operation goes through even if the row does not exist."""
        deleteVectorDBTableItemCQL = SimpleStatement(_deleteVectorDBTableItemCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
        ))
        self._executeCQL(deleteVectorDBTableItemCQL, (document_id, ))

    def search(self, embedding_vector, topK, maxRowsToRetrieve, metric, metricThreshold):
        # get rows by ANN
        rows = list(self.ANNSearch(embedding_vector, maxRowsToRetrieve))
        if rows:
            # sort, cut, validate and prepare for returning (if any)
            #
            # evaluate metric
            distanceFunction = distanceMetricsMap[metric]
            rowEmbeddings = [
                row.embedding_vector
                for row in rows
            ]
            # enrich with their metric score
            rowsWithMetric = list(zip(
                distanceFunction[0](rowEmbeddings, embedding_vector),
                rows,
            ))
            # sort rows by metric score. First handle metric/threshold
            if metricThreshold is not None:
                if distanceFunction[1]:
                    def _thresholder(mtx, thr): return mtx >= thr
                else:
                    def _thresholder(mtx, thr): return mtx <= thr
            else:
                # no hits are discarded
                def _thresholder(mtx, thr): return True
            #
            sortedPassingWinners = sorted(
                (
                    pair
                    for pair in rowsWithMetric
                    if _thresholder(pair[0], metricThreshold)
                ),
                key=itemgetter(0),
                reverse=distanceFunction[1],
            )[:topK]
            # we discard the scores and return an iterable of hits (as JSONs)
            return [
                VectorDBTable._jsonifyHit(hit, distance=distance)
                for distance, hit in sortedPassingWinners
            ]
        else:
            return []

    @staticmethod
    def _jsonifyHit(hit, distance):
        if distance is not None:
            distDict = {'distance': distance}
        else:
            distDict = {}
        return {
            **{
                'document_id': hit.document_id,
                'metadata': json.loads(hit.metadata_blob),
                'document': hit.document,
                'embedding_vector': hit.embedding_vector,
            },
            **distDict,
        }

    def clear(self):
        truncateVectorDBTableCQL = SimpleStatement(_truncateVectorDBTableCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
        ))
        self._executeCQL(truncateVectorDBTableCQL, tuple())

    def _createTable(self):
        createVectorDBTableCQL = SimpleStatement(_createVectorDBTableCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
            idType='UUID' if self.autoID else 'TEXT',
            embeddingDimension=self.embeddingDimension,
        ))
        self._executeCQL(createVectorDBTableCQL, tuple())

    def _executeCQL(self, statement, params):
        return self.session.execute(statement, params)



"""
management of "history" of stored blobs, grouped
by some 'session id'. Overwrites are not supported by design.
"""

from cassandra.query import SimpleStatement

_createTableCQLTemplate = """
CREATE TABLE IF NOT EXISTS {keyspace}.{tableName} (
    session_id TEXT,
    blob_id TIMEUUID,
    blob TEXT,
    PRIMARY KEY (( session_id ) , blob_id )
) WITH CLUSTERING ORDER BY (blob_id ASC);
"""
_getSessionBlobsCQLTemplate = """
SELECT blob
    FROM {keyspace}.{tableName}
WHERE session_id=%s;
"""
_storeSessionBlobCQLTemplate = """
INSERT INTO {keyspace}.{tableName} (
    session_id,
    blob_id,
    blob
) VALUES (
    %s,
    now(),
    %s
){ttlSpec};
"""
_clearSessionCQLTemplate = """
DELETE FROM {keyspace}.{tableName} WHERE session_id = %s;
"""


class StoredBlobHistory():

    def __init__(self, session, keyspace, tableName):
        self.session = session
        self.keyspace = keyspace
        self.tableName = tableName
        # Schema creation, if needed
        createTableCQL = SimpleStatement(_createTableCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
        ))
        session.execute(createTableCQL)

    def store(self, sessionId, blob, ttlSeconds):
        if ttlSeconds:
            ttlSpec = f' USING TTL {ttlSeconds}'
        else:
            ttlSpec = ''
        #
        storeSessionBlobCQL = SimpleStatement(_storeSessionBlobCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
            ttlSpec=ttlSpec,
        ))
        self.session.execute(
            storeSessionBlobCQL,
            (
                sessionId,
                blob,
            )
        )

    def retrieve(self, sessionId, maxCount=None):
        pass
        getSessionBlobsCQL = SimpleStatement(_getSessionBlobsCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
        ))
        blobRows = self.session.execute(
            getSessionBlobsCQL,
            (sessionId, )
        )
        return (
            row.blob
            for row in blobRows
        )

    def clearSessionId(self, sessionId):
        pass
        clearSessionCQL = SimpleStatement(_clearSessionCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
        ))
        self.session.execute(clearSessionCQL, (sessionId, ))



from typing import List, Any, Dict, Callable


def _table_primary_key_columns(session, keyspace, tableName) -> List[str]:
    table = session.cluster.metadata.keyspaces[keyspace].tables[tableName]
    return [
        col.name for col in table.partition_key
    ] + [
        col.name for col in table.clustering_key
    ]



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

setup(
    name='cassio',
    version='0.0.1',
    author='Stefano Lottini',
    author_email='stefano.lottini@datastax.com',
    package_dir={"": "src"},
    packages=find_packages(where='src'),
    # entry_points={
    #     "console_scripts": [
    #         "clothesline=clothesline:main",
    #     ],
    # },
    url='https://github.com/hemidactylus/cassio',
    license='LICENSE.txt',
    description='A framework-agnostic Python library to seamlessly integrate Apache Cassandra with ML/LLM/genAI workloads.',
    long_description=(here / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    install_requires=[
        "numpy>=1.0",
        "cassandra-driver>=3.24.0",
        #
        # Experimental support for VectorSimilaritySearch. Cannot be used if publishing to PyPI.
        # git+https://github.com/datastax/python-driver.git@cep-vsearch#egg=cassandra-driver
        # "cassandra-driver @ git+https://github.com/datastax/python-driver.git@cep-vsearch",
    ],
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        #
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="cassandra, ai, llm",
)



"""
handling of key-value storage on a Cassandra table.
One row per partition, serializes a multiple partition key into a string
"""

from typing import Union, List, Any

from cassandra.cluster import Session

# CQL templates
_createTableCQLTemplate = """
CREATE TABLE IF NOT EXISTS {keyspace}.{tableName} (
    key_desc TEXT,
    cache_key TEXT,
    cache_value TEXT,
    PRIMARY KEY (( key_desc, cache_key ))
);
"""
_getCachedItemCQLTemplate = """
SELECT cache_value
    FROM {keyspace}.{tableName}
WHERE key_desc=%s
    AND cache_key=%s;
"""
_deleteCachedItemCQLTemplate = """
DELETE FROM {keyspace}.{tableName}
WHERE key_desc=%s
    AND cache_key=%s;
"""
_storeCachedItemCQLTemplate = """
INSERT INTO {keyspace}.{tableName} (
    key_desc,
    cache_key,
    cache_value
) VALUES (
    %s,
    %s,
    %s
){ttlSpec};
"""
_truncateTableCQLTemplate = """
TRUNCATE TABLE {keyspace}.{tableName};
"""


class KVCache():

    def __init__(self, session: Session, keyspace: str, tableName: str, keys: List[Any]):
        self.session = session
        self.keyspace = keyspace
        self.tableName = tableName
        self.keys = keys
        self.keyDesc = '/'.join(self.keys)
        # Schema creation, if needed
        createTableCQL = _createTableCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
        )
        session.execute(createTableCQL)

    def clear(self):
        truncateTableCQL = _truncateTableCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
        )
        self.session.execute(
            truncateTableCQL
        )

    def put(self, keyDict, cacheValue, ttlSeconds):
        if ttlSeconds:
            ttlSpec = f' USING TTL {ttlSeconds}'
        else:
            ttlSpec = ''
        cacheKey = self._serializeKey([
            keyDict[k]
            for k in self.keys
        ])
        storeCachedItemCQL = _storeCachedItemCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
            ttlSpec=ttlSpec,
        )
        self.session.execute(
            storeCachedItemCQL,
            (
                self.keyDesc,
                cacheKey,
                cacheValue,
            ),
        )

    def get(self, keyDict) -> Union[None, str]:
        cacheKey = self._serializeKey([
            keyDict[k]
            for k in self.keys
        ])
        getCachedItemCQL = _getCachedItemCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
        )
        foundRow = self.session.execute(
            getCachedItemCQL,
            (self.keyDesc, cacheKey),
        ).one()
        if foundRow:
            return foundRow.cache_value
        else:
            return None

    def delete(self, keyDict) -> None:
        """ Will not complain if the row does not exist. """
        cacheKey = self._serializeKey([
            keyDict[k]
            for k in self.keys
        ])
        deleteCachedItemCQL = _deleteCachedItemCQLTemplate.format(
            keyspace=self.keyspace,
            tableName=self.tableName,
        )
        self.session.execute(
            deleteCachedItemCQL,
            (self.keyDesc, cacheKey),
        )

    def _serializeKey(self, keys: List[str]):
        return str(keys)



from typing import List

import numpy as np

# distance definitions. These all work batched in the first argument.
def distance_dotProduct(embeddingVectors: List[List[float]], referenceEmbeddingVector: List[float]) -> List[float]:
        """
        Given a list [emb_i] and a reference rEmb vector,
        return a list [distance_i] where each distance is
            distance_i = distance(emb_i, rEmb)
        At the moment only the dot product is supported
        (which for unitary vectors is the cosine difference).

        Not particularly optimized.
        """
        v1s = np.array(embeddingVectors, dtype=float)
        v2 = np.array(referenceEmbeddingVector, dtype=float)
        return list(np.dot(
            v1s,
            v2.T,
        ))


def distance_cosDifference(embeddingVectors: List[List[float]], referenceEmbeddingVector: List[float]) -> List[float]:
    v1s = np.array(embeddingVectors, dtype=float)
    v2 = np.array(referenceEmbeddingVector, dtype=float)
    return list(np.dot(
        v1s,
        v2.T,
    ) / (
        np.linalg.norm(v1s, axis=1)
        * np.linalg.norm(v2)
    ))


def distance_L1(embeddingVectors: List[List[float]], referenceEmbeddingVector: List[float]) -> List[float]:
        v1s = np.array(embeddingVectors, dtype=float)
        v2 = np.array(referenceEmbeddingVector, dtype=float)
        return list(np.linalg.norm(v1s - v2, axis=1, ord=1))


def distance_L2(embeddingVectors: List[List[float]], referenceEmbeddingVector: List[float]) -> List[float]:
        v1s = np.array(embeddingVectors, dtype=float)
        v2 = np.array(referenceEmbeddingVector, dtype=float)
        return list(np.linalg.norm(v1s - v2, axis=1, ord=2))


def distance_max(embeddingVectors: List[List[float]], referenceEmbeddingVector: List[float]) -> List[float]:
        v1s = np.array(embeddingVectors, dtype=float)
        v2 = np.array(referenceEmbeddingVector, dtype=float)
        return list(np.linalg.norm(v1s - v2, axis=1, ord=np.inf))


# The tuple is:
#   (
#       function,
#       sorting 'reverse' argument, nearest-to-farthest
#   )
# (i.e. True means that:
#     - in that metric higher is closer and that
#     - cutoff should be metric > threshold)
distanceMetricsMap = {
    'cos': (
        distance_cosDifference,
        True,
    ),
    'dot': (
        distance_dotProduct,
        True,
    ),
    'l1': (
        distance_L1,
        False,
    ),
    'l2': (
        distance_L2,
        False,
    ),
    'max': (
        distance_max,
        False,
    ),
}



"""
The global settings singleton
"""

class Globals():

    def __init__(self):
        self._experimental_vector_search = False

    @property
    def experimentalVectorSearch(self):
        return self._experimental_vector_search

    def enableExperimentalVectorSearch(self):
        self._experimental_vector_search = True

globals = Globals()
